/*
 * Catalyst Socket A Patch
 * Copyright (c) 2008, 2009 Axel Gembe
 * All rights reserved.
 *
 * $Id: csapatch.cpp 78 2009-01-04 19:46:23Z ago $
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * This tool patches an ati3duag.dll or atiumdag.dll for Socket A motherboards
 * with > 1900 MHz.
 *
 * Compile with Visual Studio 2005
 */

/* System includes */
#include <stdio.h>
#include <windows.h>
#include <imagehlp.h>
#pragma comment(lib, "imagehlp.lib")

/* Current version */
#define VERSION "1.0.1"

/* x86 NOP Opcode (0x90) */
#define NOP 0x90

/* Error exit code */
#define ERREXIT(x, y) ret = x; \
	system("pause"); \
	goto y

/* Function pointer to IsWow64Process */
typedef BOOL (WINAPI *pfnIsWow64Process)(HANDLE, PBOOL);

/*
 * This data is being searched inside the DLL file. Its basically a compare for
 * the 1900 MHz and a conditional jump. The jump gets replaced with NOP (0x90)
 * instructions so it never does the jump.
 */
const unsigned char needle[] = {
	0x6C, 0x07, 0x00, 0x00, /* (cmp ecx/edx, )76Ch */
	0x77, 0x2F              /* ja short +2Fh */
};

/* Finds a needle in a haystack. */
void *find(const void *dn, const void *dh, const size_t lenn, const size_t lenh)
{
	unsigned char *p = (unsigned char*)dh;
	unsigned char *pe = p + lenh - lenn;

	/* As long as we havent reached the end of the haystack */
	while (p <= pe) {
		/* Check if the first byte matches and if it does do a memcmp for the full length */
		if (*p == *((unsigned char*)dn) && memcmp(p, dn, lenn) == 0)
			return p;
		/* Increment haystack offset */
		p++;
	}

	return NULL;
}

/* Returns true if this process is running on WOW64 */
bool IsWow64()
{
	static pfnIsWow64Process fn = NULL;
	BOOL ret;

	/* Get pointer to IsWow64Process (older Windows versions don't have this */
	if (fn == NULL) {
		fn = (pfnIsWow64Process)GetProcAddress(
			GetModuleHandle("kernel32.dll"), "IsWow64Process");
		if (fn == NULL)
			return false;
	}

	/* Determine if we are running on WOW64 */
	if (fn(GetCurrentProcess(), &ret) == 0)
		return false;

	return (ret == TRUE) ? true : false;
}

int main(int argc, char **argv)
{
	/* Local variables */
	int ret = 0;
	HANDLE inFile, inMap;
	unsigned char *in, *patchLoc;
	DWORD inLen, inSum, inSumNew;
	PIMAGE_NT_HEADERS inPeHdr;
	bool patched = false;
	char inName[1024], bakName[1024];
	OSVERSIONINFO osVer;

	printf("Catalyst Socket A patch version %s\n\n", VERSION);

#ifndef _DEBUG
	/* Check for 64-bit emulation, disallow it */
	if (IsWow64()) {
		fprintf(stderr, "ERROR: This patch is for 32-bit windows only!\n");
		system("pause");
		return 1;
	}

	/* Check for safe mode, disallow otherwise */
	if (GetSystemMetrics(SM_CLEANBOOT) == 0) {
		fprintf(stderr, "ERROR: This patch must be run in safe mode!\n");
		system("pause");
		return 1;
	}
#endif /* _DEBUG */

	if (argc == 2) {
		/* Copy over the user-specified filename */
		strcpy_s(inName, sizeof(inName), argv[1]);
	} else if (argc > 2) {
		/* Print help */
		fprintf(stderr, "Usage: %s [file]\n", argv[0]);
		fprintf(stderr, "By default, the file from %SystemRoot%\\system32 will be patched.\n");
		ERREXIT(1, exit);
	} else {
		/* Get the windows directory */
		if (GetWindowsDirectory(inName, sizeof(inName)) == 0) {
			fprintf(stderr, "ERROR: Failed to get the windows directory, specify an input file as an argument please!\n");
			ERREXIT(1, exit);
		}

		/* Append system32 */
		strcat_s(inName, sizeof(inName), "\\system32\\");

		/* Get operating system version */
		osVer.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
		GetVersionEx(&osVer);

		/* Use atiumdag.dll on Vista and ati3duag.dll on older Versions */
		strcat_s(inName, sizeof(inName),
			(osVer.dwMajorVersion >= 6) ? "atiumdag.dll" : "ati3duag.dll");
	}

	/* Make a backup of the original file */
	strcpy_s(bakName, sizeof(bakName), inName);
	strcat_s(bakName, sizeof(bakName), ".bak");
	CopyFile(inName, bakName, FALSE);

	/* Open the file */
	inFile = CreateFile(inName, GENERIC_READ | GENERIC_WRITE, 0, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (inFile == INVALID_HANDLE_VALUE) {
		fprintf(stderr, "ERROR: Failed to open input file \"%s\"!\n", inName);
		ERREXIT(1, exit);
	}

	/* Get the file size */
	inLen = GetFileSize(inFile, NULL);

	/* Create file mapping */
	inMap = CreateFileMapping(inFile, NULL, PAGE_READWRITE, 0, 0, NULL);
	if (inMap == NULL) {
		fprintf(stderr, "ERROR: Failed to create mapping of input file!\n");
		ERREXIT(1, exit_file_open);
	}

	/* Map file to memory */
	in = (unsigned char*)MapViewOfFile(inMap, FILE_MAP_ALL_ACCESS, 0, 0, 0);
	if (in == NULL) {
		fprintf(stderr, "ERROR: Failed to map input file to memory!\n");
		ERREXIT(1, exit_file_mapping);
	}

	/* Find the location to patch */
	patchLoc = (unsigned char*)find(needle, in, sizeof(needle), inLen);
	/* Loop as long as there are patch locations found */
	while (patchLoc != 0) {
		/* Patch offset 4 & 5 (the relative short jump) with NOPs */
		patchLoc[4] = patchLoc[5] = NOP;

		/* Set status flag to success */
		patched = true;

		/* Find the next location to patch */
		patchLoc = (unsigned char*)find(needle, in, sizeof(needle), inLen);
	}

	/* If we didn't find any patch location */
	if (!patched) {
		fprintf(stderr, "ERROR: Failed to find the patch location, is the file patched already?\n");
		ERREXIT(1, exit_file_mapped);
	}

	/* Recompute the checksum */
	inPeHdr = CheckSumMappedFile(in, inLen, &inSum, &inSumNew);
	if (inPeHdr == NULL) {
		fprintf(stderr, "ERROR: Failed to checksum input file in memory!\n");
		ERREXIT(1, exit_file_mapped);
	}

	/* Write the new checksum to the file. */
	inPeHdr->OptionalHeader.CheckSum = inSumNew;

	/* File was mapped to memory */
exit_file_mapped:
	UnmapViewOfFile((void*)in);

	/* File mapping was created */
exit_file_mapping:
	CloseHandle(inMap);

	/* File was opened */
exit_file_open:
	CloseHandle(inFile);

exit:
	/* Notify the user that patching was successful */
	if (ret == 0) {
		printf("Congratulations! Patching was successful.\n");
		system("pause");
	}

	/* Return! :) */
	return ret;
}
